# -------------------------------------------------------------------------- #
# Script : danstonchat.tcl                                                    
# Auteur : ealexp
# Version : 1.1
# Date de création : 16 août 2010
# Date de dernière modification : 31 octobre 2010
# -------------------------------------------------------------------------- #

# -------------------------------------------------------------------------- #
# Prérequis : 
# Ce script nécéssite le package http pour fonctionner.
# Tcl 8.5 est également requis.
# -------------------------------------------------------------------------- #

# -------------------------------------------------------------------------- #
# Activation : 
# Pour activer le script, il faut taper .chanset #salon +dtc_script en 
# partyline.
# Aucune configuration n'est nécéssaire.
# -------------------------------------------------------------------------- #

# -------------------------------------------------------------------------- #
# Utilisation : 
# Ce script permet d'afficher des citations du site http://danstonchat.com
#
# La commande !dtc peut être utilisée de plusieurs façons : 
#
# 1) !dtc
#   Affiche une citation aléatoire.
#
# 2) !dtc <numéro>
#   Affiche la citation n°<numéro>.
#
# 3) !dtc <chaîne>
#   Fait une recherche sur <chaîne>.
#
# 4) !dtc last
#   Affiche la dernière citation parue sur le site.
# -------------------------------------------------------------------------- #

if {[catch {package require http}]} {
	putlog {[DTC] Vous avez besoin du package http pour utiliser ce script.}
	return
}

if {[catch {package require Tcl 8.5}]} {
	putlog {[DTC] Vous avez besoin de Tcl en version 8.5 au moins pour utiliser ce script.}
}

if {[info commands ::dtc::uninstall] eq "::dtc::uninstall"} {
	::dtc::uninstall
}

namespace eval ::dtc {
	#######################
	#                     #
	#    CONFIGURATION    #
	#                     #
	#######################

	########################################
	# COMMANDES PUBLIQUES ET AUTORISATIONS #
	########################################

	# Nom des commandes publiques pour afficher une citation.
	# Valeur par défaut : {!bashfr !dtc}
	variable command_names {!bashfr !dtc}

	# Flags de la commande
	# Valeur par défaut : "-|-"
	variable command_flags "-|-"
	
	###################
	# PARAMÈTRES HTTP #
	###################

	# Adresse pour récupérer les dernières citations.
	# Valeur par défaut : "http://danstonchat.com/latest.html"
	variable url_latest "http://danstonchat.com/latest.html"

	# Adresse pour récupérer une citation en particulier.
	# %d sera remplacé par le numéro de la citation.
	# Valeur par défaut : "http://danstonchat.com/%d.html"
	variable url_quote "http://danstonchat.com/%d.html"

	# Adresse pour récupérer une citation aléatoire.
	# Valeur par défaut : "http://danstonchat/random.html"
	variable url_random "http://danstonchat.com/random.html"

	# Adresse pour la recherche
	# Valeur par défaut : "http://danstonchat.com/search"
	variable url_search "http://danstonchat.com/search"
	
	# User-Agent
	# Valeur par défaut : "Opera/9.80 (X11; Linux i686; U; fr) Presto/2.6.30 Version/10.61"
	variable user_agent "Opera/9.80 (X11; Linux i686; U; fr) Presto/2.6.30 Version/10.61"

	
	#############
	# AFFICHAGE #
	#############
	
	# Note : les couleurs sont à mettre sous la forme "<avant-plan>[,arrière-plan]".
	#        Si vous ne voulez pas de couleurs, mettez une chaîne vide ("")

	# Couleur de l'ID.
	# Valeur par défaut : "01,15"
	variable id_color "01,15"

	# Couleur de la citation.
	# Valeur par défaut : "00,17"
	variable quote_color "00,14"
		
	# Couleur des erreurs
	# Valeur par défaut : "14"
	variable error_color "14"
	
	# Maximum de lignes affichées de la citation.
	# Valeur par défaut : 15
	variable max_lines 15
	
	#############
	# ANTIFLOOD #
	#############
	
	# Antiflood activé
	# Valeur par défaut : 1
	variable antiflood 1

	# Combien de commandes sont autorisées en combien de temps ?
	# Exemple : "4:20" = 4 commandes maximum en 20 secondes, les suivantes seront ignorées.
	# Valeur par défaut : "4:20"
	variable cmdflood "4:20"
	
	# Intervalle de temps minimum entre l'affichage de 2 messages
	# avertissant que l'anti-flood a été déclenché (ne réglez pas
	# cette valeur trop bas afin de ne pas être floodé par les messages
	# d'avertissement de l'anti-flood...)
	# Valeur par défaut : 20
	variable antiflood_msg_interval 20
	
	#######################################################################
	#                                                                     #
	#    NE RIEN MODIFIER APRÈS CE CADRE SI VOUS NE CONNAISSEZ PAS TCL    #
	#                                                                     #
	#######################################################################	
	
	variable script_name "DansTonChat"
	variable script_version "1.0"
	variable debug 1
	
	proc uninstall {args} {
		putlog "Désallocation des ressources de \002$::dtc::script_name\002..."
		foreach binding [lsearch -inline -all -regexp [binds *[set ns [string range [namespace current] 2 end]]*] " \{?(::)?$ns"] {
			unbind [lindex $binding 0] [lindex $binding 1] [lindex $binding 2] [lindex $binding 4]
		}
		foreach running_timer [timers] {
			if { [string match "*[namespace current]::*" [lindex $running_timer 1]] } { killtimer [lindex $running_timer 2] }
		}
		foreach running_utimer [utimers] {
			if { [string match "*[namespace current]::*" [lindex $running_utimer 1]] } { killutimer [lindex $running_utimer 2] }
		}
		namespace delete [namespace current]
	}
	
	setudef flag dtc_script
}

# Cache.
namespace eval ::dtc::cache {
	variable cache {}
}

# Antiflood. 
namespace eval ::dtc::antiflood {
	variable current_flood {}
	variable message {}
}

# ::dtc::command
#
# Procédure principale.

proc ::dtc::command {nick host hand chan text} {
	if {![channel get $chan dtc_script]} {
		return
	}

	if {($::dtc::antiflood) && ([::dtc::antiflood::check $chan] == "flood")} {
		return
	}
	
	set arg [lindex [split $text] 0]

	if {$arg eq ""} {
		# Citation aléatoire.
		set citation [::dtc::cache::get_newest_quote] 
	} elseif {[string is integer $arg]} {
		# Citation en particulier.
		set url_complete [format $::dtc::url_quote $arg]
		set data [::dtc::geturl $url_complete]
		set citation [lindex [::dtc::parse_page $data] 0]
	} elseif {$arg eq "last"} {
		# Dernière citation parue.
		set citation [lindex [::dtc::parse_page [::dtc::geturl $::dtc::url_latest]] 0]
	} else {
		# Recherche.
		set data [::dtc::geturl $::dtc::url_search -query [::http::formatQuery s $text]]
		set citation_list [::dtc::parse_page $data]
		set citation [lindex $citation_list [rand [llength $citation_list]]]
	}

	# Gestion des erreurs.
	switch -- $citation {
		"http_error" {
			# La requête HTTP a échoué.
			puthelp "PRIVMSG $chan :[code $::dtc::error_color $chan]La connexion au site n'a pas pu être établie. Veuillez réessayer plus tard.[code endcolor $chan]"
			return
		}
		
		"parse_error" {
			# Il n'a pas été possible de récupérer des informations dans la page.
			puthelp "PRIVMSG $chan :[code $::dtc::error_color $chan]Aucune information identifiable n'a pu être récupérée. Le site rencontre peut-être un problème ou bien le layout du site a changé.[code endcolor $chan]"
			return
		}
		
		"404_error" {
			# Une page n'a pas été trouvée.
			puthelp "PRIVMSG $chan :[code $::dtc::error_color $chan]La citation n'existe pas.[code endcolor $chan]"
			return
		}
	}

	lassign $citation id quote
	
	set lines [split $quote \n]
	set counter 0
	foreach line $lines {
		# Si c'est la première ligne, on affiche aussi l'ID de la citation.
		if {$counter == 0} {
			puthelp "PRIVMSG $chan :[code $::dtc::id_color $chan]\[$id\][code endcolor $chan] [code $::dtc::quote_color $chan][string trim $line][code endcolor $chan]"
		} elseif {([llength $lines] > $::dtc::max_lines) && ($counter == ($::dtc::max_lines - 1))} {
			# Si le nombre total de lignes de la citation est supérieur au nombre total de lignes autorisés 
			# ET qu'on est à la dernière ligne qu'on doit afficher, on ajoute à la fin de la ligne "(...)"
			
			puthelp "PRIVMSG $chan :[code $::dtc::quote_color $chan]$line[code endcolor $chan]\003\00304  (...)\003"
		} elseif {([llength $lines] > $::dtc::max_lines) && ($counter == $::dtc::max_lines)} {
			# Si le nombre total de lignes de la citation est supérieur au nombre total de lignes autorisés 
			# et qu'on a déjà affiché le nombre de lignes autorisé, on affiche l'URL de la quote

			puthelp "PRIVMSG $chan :La suite ici : [format $::dtc::url_quote $id]"
			break
		} else {
			# Sinon, on affiche la ligne normalement.
			
			# Contournement le filtrage des lignes en double dans la file du serveur.

			# D'abord on regarde combien de lignes identiques à la ligne courante ont déjà été affichés.
			# Si le nombre est supérieur à 2, on ajoute le caractère \003\040 à la ligne autant de fois que nécéssaire.
			set nb [lcount [lrange $lines 0 $counter-1] *$line*]
			for {set i 2} {$i <= $nb} {incr i} {
				append line "\003\040"
			}

			puthelp "PRIVMSG $chan :[code $::dtc::quote_color $chan]$line[code endcolor $chan]"
		}
		incr counter
	}
}

# ::dtc::lcount
#
# Compte le nombre d'occurences d'un pattern dans une liste

proc ::dtc::lcount {list pattern} {
	return [llength [lsearch -all -glob $list $pattern]]
}

# ::dtc::parse_page
#
# Retourne toutes les citations présentes dans la page passée en paramètre, sauf si celle-ci contient un code d'erreur (c'est à dire que le paramètre ne contient que "404_error" ou "http_error")
# Si aucune information n'a pu être récupérée, elle retourne "parse_error".

proc ::dtc::parse_page {data} {
	if {($data == "404_error") || ($data == "http_error")} {
		return $data
	}

	foreach {dummy quote id} [regexp -all -inline {<div class="item-content">\n (.+?)\n</div>(?:.*?)<span id="(.*?)" class="item-infos">} $data] {
		lappend quote_list [list $id [::dtc::string_filter $quote]]
	}
	
	if {![info exists quote_list]} {
		return "parse_error"
	}

	return $quote_list
}

# ::dtc::cache::usable
#
# Vérifie que le cache n'est pas vide et qu'il reste des citations non-lues.

proc ::dtc::cache::usable {} {
	lassign $::dtc::cache::cache last_quote_read quote_list
	return [expr {$last_quote_read ne "" && $last_quote_read+1 != [llength $quote_list]}]
}

# ::dtc::cache::feed
#
# Vide le cache et récupére d'autres citations.

proc ::dtc::cache::feed {} {
	set quote_list [::dtc::parse_page [::dtc::geturl $::dtc::url_random]]
	set ::dtc::cache::cache [list -1 $quote_list]
}

# ::dtc::cache::get_newest_quote
#
# Retourne la première citation non-lue dans le cache.

proc ::dtc::cache::get_newest_quote {} {
	if {![::dtc::cache::usable]} {
		::dtc::cache::feed
	}
	
	lassign $::dtc::cache::cache last_quote_read quote_list
	lset ::dtc::cache::cache 0 [expr {$last_quote_read + 1}]
	

	return [lindex $quote_list $last_quote_read+1]
}

# ::dtc::string_filter
#
# Filtre le texte de la quote.
# Cette procédure enlève les balises <br /> et les entités HTML.

proc ::dtc::string_filter {str} {
	set str [string map -nocase {
		"&agrave;"			"à"		"&agrave;"			"à"		"&aacute;"			"á"		"&acirc;"			"â"
		"&atilde;"			"ã"		"&auml;"			"ä"		"&aring;"			"å"		"&aelig;"			"æ"
		"&ccedil;"			"ç"		"&egrave;"			"è"		"&eacute;"			"é"		"&ecirc;"			"ê"
		"&euml;"			"ë"		"&igrave;"			"ì"		"&iacute;"			"í"		"&icirc;"			"î"
		"&iuml;"			"ï"		"&eth;"				"ð"		"&ntilde;"			"ñ"		"&ograve;"			"ò"
		"&oacute;"			"ó"		"&ocirc;"			"ô"		"&otilde;"			"õ"		"&ouml;"			"ö"
		"&divide;"			"÷"		"&oslash;"			"ø"		"&ugrave;"			"ù"		"&uacute;"			"ú"
		"&ucirc;"			"û"		"&uuml;"			"ü"		"&yacute;"			"ý"		"&thorn;"			"þ"
		"&yuml;"			"ÿ"		"&quot;"			"\""	"&amp;"				"&"		"&euro;"			""
		"&oelig;"			""		"&Yuml;"			""		"&nbsp;"			" "		"&iexcl;"			"¡"
		"&cent;"			"¢"		"&pound;"			"£"		"&curren;"			"¤"		"&yen;"				"¥"
		"&brvbar;"			"¦"		"&brkbar;"			"¦"		"&sect;"			"§"		"&uml;"				"¨"
		"&die;"				"¨"		"&copy;"			"©"		"&ordf;"			"ª"		"&laquo;"			"«"
		"&not;"				"¬"		"&shy;"				"­-"	"&reg;"				"®"		"&macr;"			"¯"
		"&hibar;"			"¯"		"&deg;"				"°"		"&plusmn;"			"±"		"&sup2;"			"²"
		"&sup3;"			"³"		"&acute;"			"´"		"&micro;"			"µ"		"&para;"			"¶"
		"&middot;"			"·"		"&cedil;"			"¸"		"&sup1;"			"¹"		"&ordm;"			"º"
		"&raquo;"			"»"		"&frac14;"			"¼"		"&frac12;"			"½"		"&frac34;"			"¾"
		"&iquest;"			"¿"		"&Agrave;"			"À"		"&Aacute;"			"Á"		"&Acirc;"			"Â"
		"&Atilde;"			"Ã"		"&Auml;"			"Ä"		"&Aring;"			"Å"		"&AElig;"			"Æ"
		"&Ccedil;"			"Ç"		"&Egrave;"			"È"		"&Eacute;"			"É"		"&Ecirc;"			"Ê"
		"&Euml;"			"Ë"		"&Igrave;"			"Ì"		"&Iacute;"			"Í"		"&Icirc;"			"Î"
		"&Iuml;"			"Ï"		"&ETH;"				"Ð"		"&Dstrok;"			"Ð"		"&Ntilde;"			"Ñ"
		"&Ograve;"			"Ò"		"&Oacute;"			"Ó"		"&Ocirc;"			"Ô"		"&Otilde;"			"Õ"
		"&Ouml;"			"Ö"		"&times;"			"×"		"&Oslash;"			"Ø"		"&Ugrave;"			"Ù"
		"&Uacute;"			"Ú"		"&Ucirc;"			"Û"		"&Uuml;"			"Ü"		"&Yacute;"			"Ý"
		"&THORN;"			"Þ"		"&szlig;"			"ß"		"\r"				""		"\t"				""
		"&#039;"			"\'"	"&#39;"				"\'"	"&gt;"				">"		"&lt;"				"<"
		"&#34;"				"\'"	"&#38;"				"&"		"#91;"				"\("	"&#92;"				"\/"
		"&#93;"				")"		"&#123;"			"("		"&#125;"			")"		"&#163;"			"£"
		"&#168;"			"¨"		"&#169;"			"©"		"&#171;"			"«"		"&#173;"			"­"
		"&#174;"			"®"		"&#180;"			"´"		"&#183;"			"·"		"&#185;"			"¹"
		"&#187;"			"»"		"&#188;"			"¼"		"&#189;"			"½"		"&#190;"			"¾"
		"&#192;"			"À"		"&#193;"			"Á"		"&#194;"			"Â"		"&#195;"			"Ã"
		"&#196;"			"Ä"		"&#197;"			"Å"		"&#198;"			"Æ"		"&#199;"			"Ç"
		"&#200;"			"È"		"&#201;"			"É"		"&#202;"			"Ê"		"&#203;"			"Ë"
		"&#204;"			"Ì"		"&#205;"			"Í"		"&#206;"			"Î"		"&#207;"			"Ï"
		"&#208;"			"Ð"		"&#209;"			"Ñ"		"&#210;"			"Ò"		"&#211;"			"Ó"
		"&#212;"			"Ô"		"&#213;"			"Õ"		"&#214;"			"Ö"		"&#215;"			"×"
		"&#216;"			"Ø"		"&#217;"			"Ù"		"&#218;"			"Ú"		"&#219;"			"Û"
		"&#220;"			"Ü"		"&#221;"			"Ý"		"&#222;"			"Þ"		"&#223;"			"ß"
		"&#224;"			"à"		"&#225;"			"á"		"&#226;"			"â"		"&#227;"			"ã"
		"&#228;"			"ä"		"&#229;"			"å"		"&#230;"			"æ"		"&#231;"			"ç"
		"&#232;"			"è"		"&#233;"			"é"		"&#234;"			"ê"		"&#235;"			"ë"
		"&#236;"			"ì"		"&#237;"			"í"		"&#238;"			"î"		"&#239;"			"ï"
		"&#240;"			"ð"		"&#241;"			"ñ"		"&#242;"			"ò"		"&#243;"			"ó"
		"&#244;"			"ô"		"&#245;"			"õ"		"&#246;"			"ö"		"&#247;"			"÷"
		"&#248;"			"ø"		"&#249;"			"ù"		"&#250;"			"ú"		"&#251;"			"û"
		"&#252;"			"ü"		"&#253;"			"ý"		"&#254;"			"þ"		
		
	} $str]

	regsub -all {<br />} $str " " str
	regexp {<a href="(.*?)">(.*)</a>} $str dummy dummy1 quote
	
	return [encoding convertto identity $quote]
}

# ::dtc::geturl
#
# Retourne le contenu d'une page après avoir suivi les redirections.
# Si la page n'existe pas, retourne "404_error". Si le package http n'a pas pu récupérer la page, retourne "http_error".

proc ::dtc::geturl {url args} {
	::http::config -useragent $::dtc::user_agent
	
	# On récupère l'hôte de l'URL d'origine, au cas où il ne serait plus spécifié dans la redirection.
	set original_url_host [lindex [regexp -inline {http://(.*?)/} $url] 0]
	
	# On suit au maximum 5 redirections.
	for {set i 0} {$i < 5} {incr i} {
		set token [::http::geturl $url {*}$args]
		if {![string match {30[1237]} [::http::ncode $token]]} {
			break
		}

		set meta [set ${token}(meta)]
		
		if {![dict exists $meta Location]} {
			break
		}
		set url [dict get $meta Location]

		if {[regexp -inline {http://(.*?)/} $url] eq ""} {
			set url $original_url_host$url
		}

		::http::cleanup $token
	}

	if {[::http::status $token] != "ok"} {
		set output "http_error"
	} elseif {[string match *404* [::http::code $token]]} {
		set output "404_error"
	} else {
		set output [::http::data $token]
	}

	::http::cleanup $token
	return $output
}

# ::dtc::code
#
# Retourne le code couleur spécifié, si les couleurs sont autorisées sur le salon.

proc ::dtc::code {code chan} {
	if {[string match *c* [lindex [split [getchanmode $chan]] 0]] || $code eq ""} {
		return ""
	}
	
	switch -- $code {
		"bold" {
			return \002
		}

		"underline" {
			return \037
		}

		"endcolor" {
			return \003
		}

		default {
			return \003$code
		}
	}
}

# ::dtc::antiflood::check
#
# Vérifie qu'il n'y ait pas de flood en cours.

proc ::dtc::antiflood::check {chan} {
	if {![dict exists $::dtc::antiflood::current_flood $chan]} {
		dict set ::dtc::antiflood::current_flood $chan 0
	}
	if {![dict exists $::dtc::antiflood::message $chan]} {
		dict set ::dtc::antiflood::message $chan 0
	}
	
	lassign [split $::dtc::cmdflood ":"] max_requests time

	if {[dict get $::dtc::antiflood::current_flood $chan] >= $max_requests} {
		if {[dict get $::dtc::antiflood::message $chan] == 0} {
			dict set ::dtc::antiflood::message $chan 1
			putquick "PRIVMSG $chan :[::dtc::code 7 $chan]:::[::dtc::code $::dtc::error_color $chan] Contrôle de flood activé pour cette commande : pas plus de $max_requests requête(s) toutes les $time secondes.[::dtc::code endcolor $chan]"
			utimer $::dtc::antiflood_msg_interval [list ::dtc::antiflood::reset_msg $chan]
		}
		return "flood"
	} else {
		dict incr ::dtc::antiflood::current_flood $chan
		utimer $time [list ::dtc::antiflood::close_instance $chan]
		return "no flood"
	}
}

# ::dtc::antiflood::close_instance

proc ::dtc::antiflood::close_instance {chan} {
	if {[dict get $::dtc::antiflood::current_flood $chan] > 0} {
		dict incr ::dtc::antiflood::current_flood $chan -1
	}
}

# ::dtc::antiflood::reset_msg

proc ::dtc::antiflood::reset_msg {chan} {
	dict set ::dtc::antiflood::message $chan 0
}

# ::dtc::debug

proc ::dtc::debug {args} {
	if {$::dtc::debug} {
		putlog "\[\002$::dtc::script_name\002\] [join $args]"
	}
}

foreach command $::dtc::command_names {
	bind pub $::dtc::command_flags $command ::dtc::command
}
bind evnt - prerehash ::dtc::uninstall
